#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

"""
TMD 啊啊啊啊啊啊啊啊啊
这里不是用来直接调用的啊啊啊啊啊
"""

from lib_not_dr.nuitka.reader import cli_main

if __name__ == "__main__":
    cli_main()

main = cli_main
